/************************************************************************/
/* TinyTalk: the real good stuff (network connections, main loop).	 */
/* */
/* Version 1.0 [ 1/24/90] : Initial implementation by ABR.		 */
/* 1.1 [ 1/25/90] : Changed for systems without FD_SET.	 */
/* 1.2 [ 1/26/90] : Added "quiet" and "nologin" control.	 */
/* 1.3 [ 1/27/90] : Increased max number of 'quiet' lines.	 */
/* */
/************************************************************************/

/* particular routines chosen for GB interface by Garrett van Cleef */
/* modifications for the GB interfacing copyright 1990 Garrett W. van Cleef */

#include "TinyMUD_copyright.h"

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <sys/errno.h>
#include <stdio.h>
#include <curses.h>
#include "client.h"

int             errno;

/* For BSD 4.2 systems. */
#ifndef FD_ZERO
#define OLD_BSD_FDSETS
#include "ultrix_cpt.h"
#endif
/* End of BSD 4.2 systems. */

#define REFRESH_TIME 500000	/* Microseconds */

#ifndef GB_HOST
#define GB_HOST "pooh.caltech.edu"	/* change these for the host system */
#endif

#ifndef GB_PORT
#define GB_PORT "2010"
#endif

extern char		*index(), *malloc();
extern struct hostent	*gethostbyname();
extern unsigned long 	 inet_addr();

static int      current_socket;
static int      connected;	/* Are we connected? */
hugestr         current_output;	/* Queued as messages arrive. */
static int      need_refresh;	/* Does input line need refresh? */
static int      done;		/* Are we all done? */

main()
{
	initscr();		/* set up screen for graphics */
	init_keyboard();
	client();
}


client()
{
	int             count;
	fd_set          readers, exceptions;
	struct timeval  refresh_timeout;
	struct timeval *tv;
	smallstr        host, port;

	printf("           Galactic Bloodshed Client ver %s\n",VERS);


	sprintf(host, "%s", GB_HOST);
	sprintf(port, "%s", GB_PORT);

	connected = FALSE;

	connect_to(host, port);

	*current_output = '\0';	/* No output yet. */

	done = FALSE;

	need_refresh = 0;	/* No keyboard typing yet. */

	do {			/* main loop */
		FD_ZERO(&readers);
		FD_ZERO(&exceptions);
		FD_SET(0, &readers);	/* Check standard input. */
		FD_SET(current_socket, &readers);	/* Check socket. */

		if (need_refresh) {
			refresh_timeout.tv_sec = 0;
			refresh_timeout.tv_usec = REFRESH_TIME;
			tv = &refresh_timeout;
		} else
			tv = NULL;

		count = select(current_socket + 1, &readers, NULL, &exceptions,
		    tv);
		if (count == -1) {
			if (errno != EINTR)	/* Control-Z will do this. */
				perror("select");
		} else if (count == 0) {
			do_refresh();
		} else {
			/* always check this, even after a command is issued */
			if (FD_ISSET(current_socket, &readers)) {
				handle_socket_input();
			} else if (FD_ISSET(0, &readers)) {
				/* look for conferencing */
				if (need_refresh)
					do_refresh();
				handle_keyboard_input();
			}
		}

	} while (!done);

	disconnect();
	cleanup_keyboard();
}

connect_to(host, port)		/* Try to make a connection. */
	smallstr        host, port;
{
	struct in_addr  host_address;
	struct sockaddr_in socket_address;
	int             err;

	/* register int current_socket; */

	get_host_address(host, &host_address);

	socket_address.sin_family = AF_INET;
	socket_address.sin_port = htons(atoi(port));
	bcopy(&host_address, &socket_address.sin_addr, sizeof(struct in_addr));

	current_socket = socket(AF_INET, SOCK_STREAM, 0);
	if (current_socket < 0) {
		perror("Couldn't open socket");
		die("Exiting.\n");
	}
	err = connect(current_socket, &socket_address, 
	    sizeof(struct sockaddr_in));
	if (err < 0) {
		perror("Couldn't connect to socket");
		die("Exiting.\n");
	}
	fcntl(current_socket, F_SETFL, FNDELAY);	/* Do we need this? */

	connected = TRUE;
	return;
}

get_host_address(name, addr)	/* Get a host address. */
	char           *name;
	struct in_addr *addr;
{
	struct hostent *temp;
	if (*name == '\0')
		die("No host address specified.\n");

	temp = gethostbyname(name);

	if (temp == NULL)
		die("Couldn't find host %s.\n", name);

	bcopy(temp->h_addr, addr, sizeof(struct in_addr));
}


disconnect()
{				/* Disconnect from current world. */
	if (connected)
		close(current_socket);

	connected = FALSE;
}

int 
transmit(s, l)			/* Send a message over the socket. */
	char           *s;
	int             l;
{
	int             err;

	err = send(current_socket, s, l, 0);
	if (err == -1)
		perror("send failed");

	return err;
}

int 
receive(s)
	char           *s;
{
	int             count;

	count = recv(current_socket, s, MAXSTRLEN-1, 0);
	if (count == -1) {
		if (errno == EWOULDBLOCK)
			*s = '\0';
		else {
			perror("recv failed");
		}
	} else
		s[count] = '\0';

	if (count <= 0)		/* Don't ask me. */
		done = TRUE;

	return count;
}


clear_refresh_pending()
{
	need_refresh = 0;
}

set_refresh_pending()
{
	need_refresh = 1;
}

handle_socket_input()
{
	string          blob;
	hugestr         bigblob;
	char           *place;

	receive(blob);
	strcat(current_output, blob);	/* Concatenate this bunch of input. */

	if (current_output[0] != '#' && current_output[0] != '$') {
		/* text data coming in */

		place = index(current_output, '\n');	/* Output any whole
							 * lines.  */
		if (place != NULL) {
			erase_keyboard_input();	/* Clear current line. */
			while (place != NULL) {
				*place = NULL;
				print_with_wrap(current_output);
				strcpy(bigblob, place + 1);
					/* Rest of buffer. */
				strcpy(current_output, bigblob);
					/* Copy it back to buffer. */
				place = index(current_output, '\n');
			}
		}
		if (strlen(current_output) > 0) {
			printf("%s", current_output);
		}
		if (strlen(current_output) > MAXSTRLEN) {
			/* Too long, flush it. */
			erase_keyboard_input();	/* Clear current line. */
			print_with_wrap(current_output);
		}
	} else {		/* map data coming in */
		if (current_output[0] == '#') {	/* orbit map */
			plot_orbit(current_output);
		} else if (current_output[0] == '$') {	/* planet map */
			plot_surface(current_output);
		}
	}
	*current_output = '\0';	/* Flush the buffer. */
}
