/*
 * Galactic Bloodshed, copyright (c) 1989 1990 by Robert P. Chansky, 
 * smq@ucscb.ucsc.edu, Univ of Ca Santa Cruz.
 *  You may copy, alter and distribute this game as you please as
 *    long as these notices remain intact.  No duplication for
 *    commercial purposes without permission of the author.
 *
 * Version 1.1 mods by:
 *		Garrett Van Cleef, Ph.D
 *		   Dept. of Physics
 *		 Ohio State University
 * V1.1 playtesting by:
 *		     Hanan Baddar
 *		      Ken Bloch
 *		     John Davis
 *		     Robert Jones
 *		 Menelaos Kafkaladis
 *		     Brett Lowry
 *		     Paul Murphy
 *		  Garrett van Cleef
 *		   Wolfgang Wenzel
 *                       and
 *		   tons of people
 *	  	  at U.C. Santa Cruz
 *
 *
 *	GB_server.c --
 *	  main loop
 */

#include "TinyMUD_copyright.h" 
/* Copyright for the original TinyMUD interface which was used for the networking */

#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <sys/errno.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

/*#include <time.h>	/* included by <sys/time.h> with bsd */
#include <sys/timeb.h>

#include <curses.h>
#include <sys/stat.h>
struct stat sbuf;
#include <signal.h>
#include <strings.h>
#include <pwd.h>
struct passwd *passwd;
int i;

/* GB headers */
#define EXTERN
#include "vars.h"
#include "ships.h"
#include "races.h"
#include "power.h"
#include "doturn.h"

#include "config.h"
#include "interface.h"
#include "buffers.h"

/* For BSD 4.2 systems. */
#ifndef FD_ZERO
#define OLD_BSD_FDSETS
#include "ultrix_cpt.h"
#endif

extern int	errno;
int	shutdown_flag = 0;

struct tm *update_tm; /* keeps track of the last update */
struct tm *current_tm;/* for watching for next update */
long clk;
char start_buf[128];
char update_buf[128];

static const char *connect_fail = "Connection refused.\n";
static const char *create_fail = "Connection refused.\n";
static const char *flushed_message = "<Output Flushed>\n";
static const char *shutdown_message = "Shutdown ordered by deity - Bye\n";
static const char *already_on = "Connection refused.\n";

int errno;

struct text_block {
	int			nchars;
	struct text_block	*nxt;
	char			*start;
	char			*buf;
};

struct text_queue {
    struct text_block *head;
    struct text_block **tail;
};

struct descriptor_data {
         int descriptor;
	int connected;
	int Playernum;
	char *output_prefix;
	char *output_suffix;
	int output_size;
	struct text_queue output;
	struct text_queue input;
	char *raw_input;
	char *raw_input_at;
	long last_time;
	int quota;
	struct descriptor_data *next;
	struct descriptor_data **prev;
} *descriptor_list = 0;

static int sock;
static int ndescriptors = 0;

void panic();
void process_commands();
void shovechars();
void shutdownsock();
struct descriptor_data *initializesock();
void make_nonblocking();
void freeqs();
void welcome_user();
void check_connect();
void close_sockets();
const char *addrout ();
void dump_users();
void set_signals();
struct descriptor_data *new_connection();
void parse_connect ();
void set_userstring ();
int do_command ();
char *strsave ();
int make_socket();
int queue_string();
int queue_write();
int process_output();
int process_input();
int bailout();
struct timeval update_quotas();

#define MALLOC(result, type, number) do {			\
	if (!((result) = (type *) malloc ((number) * sizeof (type))))	\
		panic("Out of memory");				\
	} while (0)

#define FREE(x) (free(x))

char *Desnames[] = {
	"ocean",
	"land",
	"mountainous",
	"gaseous",
	"ice",
	"plated",
	"error in des type!",	/* (illegal values) */
	"err in des type!"
};

int server_racedata;
#ifdef BSD4_2
int min_west;
extern char *timezone();
#endif

void main(argc, argv)
int argc;
char **argv;
{
int maim_stardata, s; 
int i, j, dummy;
#ifdef BSD4_2
char zone[16];
struct timeval tp;
struct timezone tzp;
#endif

    /* go do it */
clk = time(0);
current_tm = localtime(&clk);
#ifdef BSD4_2
/* all this because BSD 4.2 derived systems don't have
 * tm_zone in the tm struct
 */
gettimeofday(&tp, &tzp);
min_west = tzp.tz_minuteswest;
strcpy(zone, timezone(min_west, current_tm->tm_isdst));
#else
strcpy(zone, current_tm->tm_zone);
#endif

sprintf(start_buf, "Server started: %2d/%2d/%d %02d:%02d:%02d %s\n",
    current_tm->tm_mon+1, current_tm->tm_mday,
	current_tm->tm_year, current_tm->tm_hour, current_tm->tm_min,
    current_tm->tm_sec, zone);

	srandom(getpid());
	printf("***   Galactic Bloodshed ver %s ***\n\n", VERS);
        printf("      The update password is '%s'.\n", DAEMON_PASSWORD);
/* compute powers of 2 for bit operations */
	Two[0]=1;
	for(i=1;i<32;i++)
		Two[i]=2*Two[i-1];

	reload_star_data_base(); /* get star data */
	Getpower(Power); /* get power report from disk */
	Getblock(Blocks); /* get alliance block data */
for(i=1; i<=MAXPLAYERS; i++) {
	setbit(Blocks[i-1].invite, i);
	setbit(Blocks[i-1].pledge, i);
	}
	Putblock(Blocks);

Num_races = Numraces();
/* compute alliance block power */
sprintf(Power_blocks.time, "%2d/%2d/%d %02d:%02d:%02d %s", 
	current_tm->tm_mon+1, current_tm->tm_mday,
	current_tm->tm_year, current_tm->tm_hour,
	current_tm->tm_min, current_tm->tm_sec, zone);
 for (i=1; i<=Num_races; i++) {
	dummy=(Blocks[i-1].invite & Blocks[i-1].pledge);
				Power_blocks.members[i-1] = 0;
				Power_blocks.sectors_owned[i-1] = 0;
				Power_blocks.popn[i-1] = 0;
				Power_blocks.ships_owned[i-1] = 0;
				Power_blocks.planets_owned[i-1] = 0;
				Power_blocks.resource[i-1] = 0;
				Power_blocks.fuel[i-1] = 0;
				Power_blocks.destruct[i-1] = 0;
				Power_blocks.VPs[i-1] = 0;
	for(j=1; j<=Num_races; j++) 
			if(isset(dummy, j)){
				Power_blocks.members[i-1] += 1;
				Power_blocks.sectors_owned[i-1] += Power[j-1].sectors_owned;
				Power_blocks.popn[i-1] += Power[j-1].popn;
				Power_blocks.ships_owned[i-1] += Power[j-1].ships_owned;
				Power_blocks.planets_owned[i-1] += Power[j-1].planets_owned;
				Power_blocks.resource[i-1] += Power[j-1].resource;
				Power_blocks.fuel[i-1] += Power[j-1].fuel;
				Power_blocks.destruct[i-1] += Power[j-1].destruct;
				Power_blocks.VPs[i-1] += Blocks[j-1].VPs;
			}
	}



    shovechars (argc==1 ? GB_PORT : atoi (argv[1]) );
    close_sockets ();
printf("Going down.\n");
    exit (0);
}


int notify(Playernum, msg)
int Playernum;
char *msg;
{
    struct descriptor_data *d;

    for (d = descriptor_list; d; d = d->next) {
	if (d->connected && d->Playernum == Playernum) {
	    queue_string (d, msg);
	return(1);
	}
    }
return(0);
}

struct timeval timeval_sub(now, then)
struct timeval now;
struct timeval then;
{
    now.tv_sec -= then.tv_sec;
    now.tv_usec -= then.tv_usec;
    if (now.tv_usec < 0) {
	now.tv_usec += 1000000;
	now.tv_sec--;
    }
    return now;
}

int msec_diff(now,  then)
struct timeval now;
struct timeval then;
{
    return ((now.tv_sec - then.tv_sec) * 1000
	    + (now.tv_usec - then.tv_usec) / 1000);
}

struct timeval msec_add(t, x)
struct timeval t;
int x;
{
    t.tv_sec += x / 1000;
    t.tv_usec += (x % 1000) * 1000;
    if (t.tv_usec >= 1000000) {
	t.tv_sec += t.tv_usec / 1000000;
	t.tv_usec = t.tv_usec % 1000000;
    }
    return t;
}

void shovechars(port)
int port;
{
    fd_set input_set, output_set;
    long now;
    struct timeval last_slice, current_time;
    struct timeval next_slice;
    struct timeval timeout, slice_timeout;
    int maxd;
    struct descriptor_data *d, *dnext;
    struct descriptor_data *newd;
    int avail_descriptors;

    sock = make_socket (port);
    maxd = sock+1;
    gettimeofday(&last_slice, (struct timezone *) 0);

    avail_descriptors = getdtablesize() - 4;
    
    while (shutdown_flag == 0) {
	gettimeofday(&current_time, (struct timezone *) 0);
	last_slice = update_quotas(last_slice, current_time);

	process_commands();

	if (shutdown_flag)
	    break;
	timeout.tv_sec = 1000;
	timeout.tv_usec = 0;
	next_slice = msec_add (last_slice, COMMAND_TIME_MSEC);
	slice_timeout = timeval_sub (next_slice, current_time);
	
	FD_ZERO (&input_set);
	FD_ZERO (&output_set);
	if (ndescriptors < avail_descriptors)
	    FD_SET (sock, &input_set);
	for (d = descriptor_list; d; d=d->next) {
	    if (d->input.head)
		timeout = slice_timeout;
	    else
		FD_SET (d->descriptor, &input_set);
	    if (d->output.head)
		FD_SET (d->descriptor, &output_set);
	}

	if (select (maxd, &input_set, &output_set,
		    (fd_set *) 0, &timeout) < 0) {
	    if (errno != EINTR) {
		perror ("select");
		return;
	    }
	} else {
	    (void) time (&now);
	    if (FD_ISSET (sock, &input_set)) {
		if (!(newd = new_connection (sock))) {
		    if (errno != EINTR && errno != EMFILE) {
			perror ("new_connection");
			return;
		    }
		} else {
		if (newd->descriptor >= maxd)
		    maxd = newd->descriptor + 1;
		}
	    }
	    for (d = descriptor_list; d; d = dnext) {
		dnext = d->next;
		if (FD_ISSET (d->descriptor, &input_set)) {
			d->last_time = now;
			if (!process_input (d)) {
			    shutdownsock (d);
			    continue;
			}
		}
		if (FD_ISSET (d->descriptor, &output_set)) {
		    if (!process_output (d)) {
			shutdownsock (d);
		    }
		}
	    }
	}
    }
}

int make_socket(port)
int port;
{
    int s;
    struct sockaddr_in server;
    int opt;

    s = socket (AF_INET, SOCK_STREAM, 0);
    if (s < 0) {
	perror ("creating stream socket");
	exit (3);
    }
    opt = 1;
    if (setsockopt (s, SOL_SOCKET, SO_REUSEADDR,
		    (char *) &opt, sizeof (opt)) < 0) {
	perror ("setsockopt");
	exit (1);
    }
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_port = htons (port);
    if (bind (s, (struct sockaddr *) & server, sizeof (server))) {
	perror ("binding stream socket");
	close (s);
	exit (4);
    }
    listen (s, 5);
    return s;
}

struct timeval update_quotas(last, current)
struct timeval last;
struct timeval current;
{
    int nslices;
    struct descriptor_data *d;

    nslices = msec_diff (current, last) / COMMAND_TIME_MSEC;

    if (nslices > 0) {
	for (d = descriptor_list; d; d = d -> next) {
	    d -> quota += COMMANDS_PER_TIME * nslices;
	    if (d -> quota > COMMAND_BURST_SIZE)
		d -> quota = COMMAND_BURST_SIZE;
	}
    }
    return msec_add (last, nslices * COMMAND_TIME_MSEC);
}

struct descriptor_data *new_connection(sock)
int sock;
{
    int newsock;
    struct sockaddr_in addr;
    int addr_len;

    addr_len = sizeof (addr);
    newsock = accept (sock, (struct sockaddr *) & addr, &addr_len);
    if (newsock <= 0) {
	return 0;
    } else {
	fprintf (stderr, "ACCEPT from %s(%d) on descriptor %d\n", addrout (ntohl (addr.sin_addr.s_addr)), ntohs (addr.sin_port), newsock);
	return initializesock (newsock);
    }
}

const char *addrout(a)
long a;
{
    static char outbuf[1024];

    sprintf (outbuf, "%d.%d.%d.%d", (a >> 24) & 0xff, (a >> 16) & 0xff,
	     (a >> 8) & 0xff, a & 0xff);
    return outbuf;
}

void clearstrings(d)
struct descriptor_data *d;
{
    if (d->output_prefix) {
	free(d->output_prefix);
	d->output_prefix = 0;
    }
    if (d->output_suffix) {
	free(d->output_suffix);
	d->output_suffix = 0;
    }
}

void shutdownsock(d)
struct descriptor_data *d;
{
    if (d->connected) {
	fprintf (stderr, "DISCONNECT descriptor %d Playernum(%d)\n",
		d->descriptor, d->Playernum);
    } else {
	fprintf (stderr, "DISCONNECT descriptor %d never connected\n",
		d->descriptor);
    }
    clearstrings (d);
    shutdown (d->descriptor, 2);
    close (d->descriptor);
    freeqs (d);
    *d->prev = d->next;
    if (d->next)
        d->next->prev = d->prev;
    free(d);
    ndescriptors--;
}

struct descriptor_data *initializesock(s)
int s;
{
    struct descriptor_data *d;

    ndescriptors++;
    MALLOC(d, struct descriptor_data, 1);
    d->descriptor = s;
    d->connected = 0;
    make_nonblocking (s);
    d->output_prefix = 0;
    d->output_suffix = 0;
    d->output_size = 0;
    d->output.head = 0;
    d->output.tail = &d->output.head;
    d->input.head = 0;
    d->input.tail = &d->input.head;
    d->raw_input = 0;
    d->raw_input_at = 0;
    d->quota = COMMAND_BURST_SIZE;
    d->last_time = 0;
    if (descriptor_list)
        descriptor_list->prev = &d->next;
    d->next = descriptor_list;
    d->prev = &descriptor_list;
    descriptor_list = d;
    
    welcome_user (d);
    return d;
}

struct text_block *make_text_block(s, n)
const char *s;
int n;
{
	struct text_block *p;

	MALLOC(p, struct text_block, 1);
	MALLOC(p->buf, char, n);
	bcopy (s, p->buf, n);
	p->nchars = n;
	p->start = p->buf;
	p->nxt = 0;
	return p;
}

void free_text_block (t)
struct text_block *t;
{
		free(t->buf);
		free((char *) t);
}

void add_to_queue(q, b, n)
struct text_queue *q;
const char *b;
int n;
{
    struct text_block *p;

    if (n == 0) return;

    p = make_text_block (b, n);
    p->nxt = 0;
    *q->tail = p;
    q->tail = &p->nxt;
}

int flush_queue(q, n)
struct text_queue *q;
int n;
{
        struct text_block *p;
	int really_flushed = 0;
	
	n += strlen(flushed_message);

	while (n > 0 && (p = q->head)) {
	    n -= p->nchars;
	    really_flushed += p->nchars;
	    q->head = p->nxt;
	    free_text_block (p);
	}
	p = make_text_block(flushed_message, strlen(flushed_message));
	p->nxt = q->head;
	q->head = p;
	if (!p->nxt)
	    q->tail = &p->nxt;
	really_flushed -= p->nchars;
	return really_flushed;
}

int queue_write(d, b, n)
struct descriptor_data *d;
const char *b;
int n;
{
    int space;

    space = MAX_OUTPUT - d->output_size - n;
    if (space < 0)
        d->output_size -= flush_queue(&d->output, -space);
    add_to_queue (&d->output, b, n);
    d->output_size += n;
    return n;
}

int queue_string(d, s)
struct descriptor_data *d;
const char *s;
{
    return queue_write (d, s, strlen (s));
}

int process_output(d)
struct descriptor_data *d;
{
    struct text_block **qp, *cur;
    int cnt;

    for (qp = &d->output.head; cur = *qp;) {
	cnt = write (d->descriptor, cur -> start, cur -> nchars);
	if (cnt < 0) {
	    if (errno == EWOULDBLOCK)
		return 1;
	    return 0;
	}
	d->output_size -= cnt;
	if (cnt == cur -> nchars) {
	    if (!cur -> nxt)
		d->output.tail = qp;
	    *qp = cur -> nxt;
	    free_text_block (cur);
	    continue;		/* do not adv ptr */
	}
	cur -> nchars -= cnt;
	cur -> start += cnt;
	break;
    }
    return 1;
}

void make_nonblocking(s)
int s;
{
    if (fcntl (s, F_SETFL, FNDELAY) == -1) {
	perror ("make_nonblocking: fcntl");
	panic ("FNDELAY fcntl failed");
    }
}

void freeqs(d)
struct descriptor_data *d;
{
    struct text_block *cur, *next;

    cur = d->output.head;
    while (cur) {
	next = cur -> nxt;
	free_text_block (cur);
	cur = next;
    }
    d->output.head = 0;
    d->output.tail = &d->output.head;

    cur = d->input.head;
    while (cur) {
	next = cur -> nxt;
	free_text_block (cur);
	cur = next;
    }
    d->input.head = 0;
    d->input.tail = &d->input.head;

    if (d->raw_input)
        free(d->raw_input);
    d->raw_input = 0;
    d->raw_input_at = 0;
}

void welcome_user(d)
struct descriptor_data *d;
{
    FILE *f;
    char buf[BUFFER_LEN];
    char *p;
    
	sprintf(buf, "***   Welcome to Galactic Bloodshed ver %s ***\n", VERS);
	queue_string(d, buf);

    if((f = fopen(WELCOME_FILE, "r")) != NULL) {
	while(fgets(buf, sizeof buf, f)) {
	    for(p = buf; *p; p++) if(*p == '\n') {
		*p = '\0';
		break;
	    }
	    queue_string(d, buf);
	    queue_string(d, "\n");
	}
	fclose(f);
    }

/*    if((f = fopen(NEWS_FILE, "r")) != NULL) {
	while(fgets(buf, sizeof buf, f)) {
	    for(p = buf; *p; p++) if(*p == '\n') {
		*p = '\0';
		break;
	    }
	    queue_string(d, buf);
	    queue_string(d, "\n");
	}
	fclose(f);
    } */

}

void goodbye_user(d)
struct descriptor_data *d;
{
    write (d->descriptor, LEAVE_MESSAGE, strlen (LEAVE_MESSAGE));
}

char *strsave (s)
const char *s;
{
    char *p;

    MALLOC (p, char, strlen(s) + 1);

    if (p)
	strcpy (p, s);
    return p;
}

void save_command (d, command)
struct descriptor_data *d;
const char *command;

{
    add_to_queue (&d->input, command, strlen(command)+1);
}

int process_input (d)
struct descriptor_data *d;
{
    int got;
    char *p, *pend, *q, *qend;

    got = read (d->descriptor, buf, sizeof buf);
    if (got <= 0)
	return 0;
    if (!d->raw_input) {
	MALLOC(d->raw_input,char,MAX_COMMAND_LEN);
	d->raw_input_at = d->raw_input;
    }
    p = d->raw_input_at;
    pend = d->raw_input + MAX_COMMAND_LEN - 1;
    for (q=buf, qend = buf + got; q < qend; q++) {
	if (*q == '\n') {
	    *p = '\0';
	    if (p > d->raw_input)
		save_command (d, d->raw_input);
	    p = d->raw_input;
	} else if (p < pend && isascii (*q) && isprint (*q)) {
	    *p++ = *q;
	}
    }
    if (p > d->raw_input) {
	d->raw_input_at = p;
    } else {
	free(d->raw_input);
	d->raw_input = 0;
	d->raw_input_at = 0;
    }
    return 1;
}

void set_userstring (userstring, command)
char **userstring;
const char *command;
{
    if (*userstring) {
	free(*userstring);
	*userstring = 0;
    }
    while (*command && isascii (*command) && isspace (*command))
	command++;
    if (*command)
	*userstring = strsave (command);
}

#ifdef __STDC__
void process_commands(void)
#else
void process_commands()
#endif
{
    int nprocessed;
    struct descriptor_data *d, *dnext;
    struct text_block *t;

    do {
	nprocessed = 0;
	for (d = descriptor_list; d; d = dnext) {
	    dnext = d->next;
	    if (d -> quota > 0 && (t = d -> input.head)) {
		d -> quota--;
		nprocessed++;

		if (!do_command (d, t -> start)) {
		    shutdownsock (d);
		} else {
		    d -> input.head = t -> nxt;
		    if (!d -> input.head)
			d -> input.tail = &d -> input.head;
		    free_text_block (t);
		}

	    }
	}
    } while (nprocessed > 0);

}

int do_command (d, command)
struct descriptor_data *d;
char *command;
{
racetype *r;
char *string;
int parse_exit=0, argn=0, i;
char args[MAXARGS][COMMANDSIZE];

/* Main processing loop. When command strings are sent from the client,
	they are processed here. Responses are sent back to the client via notify.
	The client will then process the strings in whatever manner is expected. */

/* check to see if there are a few words typed out, usually for the help command */
	string=command;
	while (!parse_exit) {
	    i=0;
	    while ( !isspace(*string) && (*string!='\0') && (i<COMMANDSIZE) )
		args[argn][i++]=(*string++);
	    args[argn][i]='\0';
	    /*printf("arg %d |%s|\n",argn,args[argn]);*/
	    while ((*string)==' ') string++;
	    if ( (*string=='\0') || (argn>=MAXARGS)) parse_exit=1;
	    argn++;
	}
	for (i=argn; i<MAXARGS; i++)
		args[i][0]='\0';

   if (!strcmp (args[0], QUIT_COMMAND)) {
	if(d->connected) {

	openracedata(&server_racedata);
		  getrace(server_racedata, &r, d->Playernum);
	close_file(server_racedata);

     	sprintf(buf, "%s (%d) logged out.\n", r->name, d->Playernum);
   	 for(i=1; i<=Numraces(); i++)
		if(!Dir[i-1].gag)
              		notify(i, buf);
			}
	goodbye_user (d);
	return 0;
    } else if (!strcmp (args[0], WHO_COMMAND)) {
	dump_users (d);
    } else if (!strcmp (args[0], HELP_COMMAND)) { 
	help(d, argn, args);
    } else {
	if (d->connected) {
	    process_command (d->Playernum, command);  /* GB command parser */
	} else {
	    check_connect (d, command); /* Logs player into the game, connects
								if the password given by *command is a player's */
	    if(!d->connected) {
			goodbye_user(d);
		    } 
		else {
	    check_for_telegrams(d->Playernum);
		process_command (d->Playernum, "cs");  /* set the scope to home upon login */
			}	
	}
    }
    return 1;
}

void check_connect (d, msg)
struct descriptor_data *d;
const char *msg;
{
    char password[MAX_COMMAND_LEN];
    reg int i;
    int Playernum, Num_races;
    racetype *r;
    struct descriptor_data *d0, *dnext;
#ifdef BSD4_2
    char zone[16];
#endif

    parse_connect (msg,  password);

Num_races = Numraces();
/* daemon wants an update */
if(!strcmp(password, DAEMON_PASSWORD)) { 

for (i=1; i<=Num_races; i++)
	notify(i, "DOING UPDATE...\n");


    clk = time(0);
    update_tm=localtime(&clk);
#ifdef BSD4_2
    strcpy(zone, timezone(min_west, update_tm->tm_isdst));
#else
    strcpy(zone, update_tm->tm_zone);
#endif
sprintf(Power_blocks.time, "%2d/%2d/%d %02d:%02d:%02d %s", 
	update_tm->tm_mon+1, update_tm->tm_mday,
	update_tm->tm_year, update_tm->tm_hour,
	update_tm->tm_min, update_tm->tm_sec, zone);
sprintf(update_buf, "Last update   : %2d/%2d/%d %02d:%02d:%02d %s\n",
	update_tm->tm_mon+1, update_tm->tm_mday,
	update_tm->tm_year, update_tm->tm_hour,
	update_tm->tm_min, update_tm->tm_sec, zone);
    do_turn();
	return;
	}


    i = Getracenum(password);

	if (!i) {
	    queue_string (d, connect_fail);
	    fprintf (stderr, "FAILED CONNECT %s on descriptor %d\n",
		     password, d->descriptor);
	} else {

	openracedata(&server_racedata);
		  getrace(server_racedata, &r, i);
	close_file(server_racedata);

		  Playernum = r->Playernum;

/* check to see if this player is already connect, if so, nuke the descriptor */
	for (d0 = descriptor_list; d0; d0 = dnext) {
	    dnext = d0->next;	
		if(d0->connected && d0->Playernum == Playernum) {
			queue_string (d, already_on);
			return;
		}
	}

	    fprintf (stderr, "CONNECTED %s(%d) on descriptor %d\n",
		     r->name, Playernum, d->descriptor);
	    d->connected = 1;
	    d->Playernum = Playernum;

		sprintf(buf,"\nLogged in as %s (%d).\n",r->name,Playernum);
		notify(Playernum, buf);
     
		sprintf(buf, "%s (%d) logged on.\n", r->name, Playernum);
		for(i=1; i<=Numraces(); i++)
	          	if(i != Playernum && !Dir[i-1].gag) notify(i, buf);

	GB_time(Playernum);
	if (r->Gov_ship==0) {
		sprintf(buf,"You have no Governmental Center.  No action points will be produced\nuntil you build one and designate a capital.\n");
		notify(Playernum, buf);
		} else {
		sprintf(buf,"Government Center #%d is active.\n", r->Gov_ship);
		notify(Playernum, buf);
		}
	}
}

void parse_connect (msg, pass)
const char *msg;
char *pass;
{
    char *p;

    while (*msg && isascii(*msg) && isspace (*msg))
	msg++;
    p = pass;
    while (*msg && isascii(*msg) && !isspace (*msg))
	*p++ = *msg++;
    *p = '\0';
}

#ifdef __STDC__
void close_sockets(void)
#else
void close_sockets()
#endif
{
    struct descriptor_data *d, *dnext;

    for (d = descriptor_list; d; d = dnext) {
	dnext = d->next;
	write (d->descriptor, shutdown_message, strlen (shutdown_message));
	if (shutdown (d->descriptor, 2) < 0)
	    perror ("shutdown");
	close (d->descriptor);
    }
    close (sock);
}

#ifdef __STDC__
void emergency_shutdown(void)
#else
void emergency_shutdown()
#endif
{
	close_sockets();
}

#ifdef __STDC__
int bailout(int sig, int code, struct sigcontext *scp)
#else
int bailout(sig, code, scp)
int sig, code;
struct sigcontext *scp;
#endif
{
    char message[1024];
    
    sprintf (message, "BAILOUT: caught signal %d code %d", sig, code);
    panic(message);
    _exit (7);
    return 0;
}

#ifdef __STDC__
int dump_status(void)
#else
int dump_status()
#endif
{
    struct descriptor_data *d;
    long now;
    racetype *r;

    (void) time (&now);
    fprintf (stderr, "STATUS REPORT:\n");
openracedata(&server_racedata);
    for (d = descriptor_list; d; d = d->next) {
	if (d->connected) {
	    getrace(server_racedata, &r, d->Playernum);
	    fprintf (stderr, "PLAYING descriptor %d Playernum %s(%d)",
		     d->descriptor, r->name, d->Playernum);

	    if (d->last_time)
		fprintf (stderr, " idle %d seconds\n",
			 now - d->last_time);
	    else
		fprintf (stderr, " never used\n");
	} else {
	    fprintf (stderr, "CONNECTING descriptor %d", d->descriptor);
	    if (d->last_time)
		fprintf (stderr, " idle %d seconds\n",
			 now - d->last_time);
	    else
		fprintf (stderr, " never used\n");
	}
    }
close_file(server_racedata); 
   return 0;
}

void dump_users(e)
struct descriptor_data *e;
{
    struct descriptor_data *d;
    long now;
    racetype *r;

    (void) time (&now);
    queue_string (e, "Current Players:\n");
openracedata(&server_racedata);
    for (d = descriptor_list; d; d = d->next) {
	if (d->connected) {
	    getrace(server_racedata, &r, d->Playernum);

	    if (d->last_time)
		sprintf (buf, "%20s (%3d)     seconds idle %d\n", r->name, 
			d->Playernum, now - d->last_time);
	    else
		sprintf (buf, "%20s (%3d)     idle forever\n", r->name, d->Playernum);
	    queue_string (e, buf);
	}
    }
close_file(server_racedata);
}

void process_command(Playernum, command) 
/* locate and parse the command for processing */
int Playernum;
char *command;
{
int i,argn=0;
boolean parse_exit=0;
char args[MAXARGS][COMMANDSIZE];
char *string, *announcement, *dummy;

openracedata(&server_racedata);
free(Race);
getrace(server_racedata, &Race, Playernum);
close_file(server_racedata);

	string=command;	/* make pointers the same */

	while (!parse_exit) {
	    i=0;
	    while ( !isspace(*string) && (*string!='\0') && (i<COMMANDSIZE) )
		args[argn][i++]=(*string++);
	    args[argn][i]='\0';
	    /*printf("arg %d |%s|\n",argn,args[argn]);*/
	    while ((*string)==' ') string++;
	    if ( (*string=='\0') || (argn>=MAXARGS)) parse_exit=1;
	    argn++;
	}
	for (i=argn; i<MAXARGS; i++)
		args[i][0]='\0';

	if (args[0][0]=='\0')   /* just pressed return */
		return;	/* go back and wait again */
	i=strlen(args[0]);

/*determine which routine to go to and what action to take */
/* target routine is specified by the first substring, other options 
	are the substrings which follow */

        if (!strncmp(args[0],"announce",i))
			announce(Playernum,0,argn,args, 0);
         else if (!strncmp(args[0],"allocate",i)) allocateAPs(Playernum,0, argn, args);
         else if (!strncmp(args[0],"assault",i)) land(Playernum,1,argn,args,1);
	else if (!strncmp(args[0], "autoreport",i)) autoreport(Playernum, 0, argn, args);
	else if (!strncmp(args[0],"build",i)) build(Playernum,1,argn,args);
	else if (!strncmp(args[0], "broadcast", i)) announce(Playernum, 5*(!Race->God), argn, args, 1);
	else if (!strncmp(args[0], "block", i)) block(Playernum, 0, argn, args);
         else if (!strncmp(args[0],"capitol",i)) capitol(Playernum,50,argn,args);
         else if (!strncmp(args[0],"capture",i)) capture(Playernum,1,argn,args);
	else if (!strncmp(args[0], "center", i)) center(Playernum, 0, argn, args);
	else if (!strncmp(args[0],"cs",i)) cs(Playernum,0,argn,args);
	else if (!strncmp(args[0],"declare",i)) declare(Playernum,1,argn,args);
	else if (!strncmp(args[0],"dock",i)) land(Playernum,0,argn,args,0);
	else if (!strncmp(args[0],"enslave",i)) enslave(Playernum,2,argn,args);
	else if (!strncmp(args[0],"examine",i)) examine(Playernum,0,argn,args);
	else if (!strncmp(args[0],"explore",i)) exploration(Playernum,0,argn,args);
	else if (!strncmp(args[0],"fire",i)) fire(Playernum,1,argn,args);
	else if (!strncmp(args[0], "gag", i)) Dir[Playernum-1].gag = !Dir[Playernum-1].gag;
	else if (!strncmp(args[0],"give",i)) give(Playernum,5,argn,args);
	else if (!strncmp(args[0], "identify", i)) whois(Playernum, 0, argn, args);
	else if (!strncmp(args[0], "invite", i)) invite(Playernum, 0, argn, args, 1);
	else if (!strncmp(args[0],"land",i)) land(Playernum,1,argn,args,0);
	else if (!strncmp(args[0],"launch",i)) launch(Playernum,1,argn,args);
	else if (!strncmp(args[0], "list", i) && Race->God) list(Playernum);
	else if (!strncmp(args[0],"load",i)) load(Playernum,0,argn,args);
	else if (!strncmp(args[0],"map",i)) map(Playernum,0,argn,args);
	else if (!strncmp(args[0],"mobilize",i)) mobilize(Playernum,1,argn,args);
	else if (!strncmp(args[0],"move",i)) move_popn(Playernum,1,argn,args);
	else if (!strncmp(args[0],"make", i)) make_mod(Playernum, 0, argn, args, 0);
	else if (!strncmp(args[0],"modify", i)) make_mod(Playernum, 0, argn, args, 1);
	else if (!strncmp(args[0],"name",i)) name(Playernum,1*(!Race->God),argn,args);
	else if (!strncmp(args[0],"object",i)) rst(Playernum,0,argn,args,4);
	else if (!strncmp(args[0],"orbit",i)) orbit(Playernum,0,argn,args);
	else if (!strncmp(args[0],"order",i)) order(Playernum,1,argn,args); 
	else if (!strncmp(args[0],"page",i)) page(Playernum, 1*(!Race->God), argn, args);
         else if (!strncmp(args[0],"colonies",i)) colonies(Playernum,0,argn,args);
	else if (!strncmp(args[0], "pledge", i)) pledge(Playernum, 0, argn, args, 1);
	else if (!strncmp(args[0],"power",i)) power(Playernum,0,argn,args);
	else if (!strncmp(args[0],"profile",i)) profile(Playernum,0,argn,args);
	else if (!strncmp(args[0],"relation",i)) relation(Playernum,0,argn,args);
	else if (!strncmp(args[0],"read",i)) read_messages(Playernum,0,argn,args);
	else if (!strncmp(args[0],"reset", i) && Race->God) reload_star_data_base();
	else if (!strncmp(args[0],"report",i)) rst(Playernum,0,argn,args,0);
	else if (!strncmp(args[0],"@@shutdown",i) && Race->God) {
		shutdown_flag = 1;
		notify(Playernum, "Doing shutdown.\n");
	} else if (!strncmp(args[0],"stock",i)) rst(Playernum,0,argn,args,1);
	else if (!strncmp(args[0],"ship",i)) rst(Playernum,0,argn,args,3);
	else if (!strncmp(args[0],"survey",i)) survey(Playernum,0,argn,args);
	else if (!strncmp(args[0], "send", i)) send_message(Playernum, 1*(!Race->God), argn, args);
	else if (!strncmp(args[0],"scrap",i)) scrap(Playernum,1,argn,args);
	else if(!strncmp(args[0],"tactical",i)) rst(Playernum,0,argn,args,2);
/*	else if (!strncmp(args[0],"telegram",i)) telegram(Playernum,0,argn,args); */
	else if(!strncmp(args[0],"technology",i)) technology(Playernum,1,argn,args);
	else if (!strncmp(args[0],"toggle",i)) toggle(Playernum,0,argn,args);
	else if(!strncmp(args[0],"toxicity",i)) toxicity(Playernum,1,argn,args);
	else if (!strncmp(args[0],"transfer",i)) transfer(Playernum,1,argn,args);
	else if (!strncmp(args[0],"unload",i)) load(Playernum,0,argn,args);
	else if (!strncmp(args[0],"undock",i)) launch(Playernum,0,argn,args);
	else if (!strncmp(args[0], "uninvite", i)) invite(Playernum, 0, argn, args, 0); 
	else if (!strncmp(args[0], "unpledge", i)) pledge(Playernum, 0, argn, args, 0); 
        else if (!strncmp(args[0],"time",i)) GB_time(Playernum);
	else if (!strncmp(args[0],"zoom",i)) zoom(Playernum,0,argn,args);
			/* only owning player can use fix */
	 else {
	       sprintf(buf,"'%s':illegal command error(%d).\n",args[0],*args[0]);
		notify(Playernum, buf);
		}

/* send the prompt to the player */
	sprintf(buf, "%s", Dir[Playernum-1].prompt);
		notify(Playernum, buf);

}

void panic(message)
const char *message;
{
    FILE *f;
    int i;

    fprintf(stderr, "PANIC: %s\n", message);

    /* turn off signals */
    for(i = 0; i < NSIG; i++) {
	signal(i, SIG_IGN);
    }

    /* shut down interface */
    emergency_shutdown();

    /* dump panic file */
}

reload_star_data_base()
{
int maim_stardata;
int s, t;

/* get star database */
	openstardata(&maim_stardata);
	  getsdata(maim_stardata,&Sdata);
	for (s=0; s<Sdata.numstars; s++) {
		getstar(maim_stardata,&(Stars[s]),s);
		}
	close_file(maim_stardata);
/* initialize zoom factors */	
	for(i=1; i<=Numraces(); i++) {
	Dir[i-1].zoom[0] = 1.0;
	Dir[i-1].zoom[1] = 0.5;
	Dir[i-1].lastx[0] = Dir[i-1].lastx[1] = 0.0;
	Dir[i-1].lasty[0] = Dir[i-1].lasty[1] = 0.0;
	}
}


str_cat(s1, s2)
char *s1,*s2;
{
	strcat(s1, s2);
}

GB_time(Playernum) /* report back the update status */
int Playernum;
{
#ifdef BSD4_2
    char zone[16];
#endif

    clk = time(0);
    current_tm=localtime(&clk);
#ifdef BSD4_2
    strcpy(zone, timezone(min_west, current_tm->tm_isdst));
#else
    strcpy(zone, current_tm->tm_zone);
#endif
sprintf(buf,"\nCurrent time  : %2d/%2d/%d %02d:%02d:%02d %s\n",
	current_tm->tm_mon+1, current_tm->tm_mday,
        current_tm->tm_year, current_tm->tm_hour,
	current_tm->tm_min, current_tm->tm_sec, zone);
notify(Playernum, buf);
if(update_tm != NULL)
   notify(Playernum, update_buf);
else
   notify(Playernum, "No updates since the server was started.\n");
    /* can happen if you have just started the game up */   
notify(Playernum, start_buf);
}

help(e, argn,args)
struct descriptor_data *e;
int argn;
char args[MAXARGS][COMMANDSIZE];
{
 char s[MAXCOMMSTRSIZE];

 if (argn==1) {
	sprintf(buf,"usage: help [topic]\n");
		queue_string(e, buf);
	sprintf(buf,"    valid topics are:\n  CONCEPTS:\n%s",
		"help     scope    actionpoints       explore    colony\n" );
		queue_string(e, buf);
	sprintf(buf,"daemon   races     blocks     quit  	     info\n");
		queue_string(e, buf);
	sprintf(buf,"\n  COMMANDS:\n");
		queue_string(e, buf);
	sprintf(buf,"cs       map      survey  orbit      zoom     autoreport\n");
		queue_string(e, buf);
	sprintf(buf,"toggle   gag	  capitol  allocate\n\n");
		queue_string(e, buf);
	sprintf(buf, "announce page    who    time    quit\n");
		queue_string(e, buf);
	sprintf(buf,"build    ship     stock   report     tactical  order     examine\n");
		queue_string(e, buf);
	sprintf(buf,"land     launch   dock    undock     load      assault   scrap\n\n");
		queue_string(e, buf);
	sprintf(buf,"assault  capture  fire    give       move       enslave\n\n");
		queue_string(e, buf);
	sprintf(buf,"declare  profile  relation    power   explore\n");
		queue_string(e, buf);
	sprintf(buf,"colonies name     mobilize technology toxicity   transfer\n");
		queue_string(e, buf);
	sprintf(buf, "make   modify\n");
		queue_string(e, buf); 

 } else {

    FILE *f;
    char file[1024];
    char *p;

    sprintf(file,"%s/%s.doc", DOCS, args[1] );
    if((f = fopen(file, "r")) != 0) {
	while(fgets(buf, sizeof buf, f)) {
	    for(p = buf; *p; p++) if(*p == '\n') {
		*p = '\0';
		break;
	    }
	strcat(buf, "\n");
		queue_string(e, buf);
	}
	fclose(f);
    } else
	queue_string(e, "Help on that subject unavailable.\n");

  }

}

check_for_telegrams(Playernum)
int Playernum;
{
sprintf(buf, "%s.%d", TELEGRAMFL, Playernum);
stat(buf, &sbuf);
if(sbuf.st_size)
notify(Playernum, "You have telegram(s) waiting. Use 'read' to read them.\n");
}
